package loggo

import (
	"fmt"
	"runtime"
	"sort"
	"strings"
	"time"
)

type Level int

const (
	NOT_SPECIFIED Level = iota
	TRACE
	DEBUG
	INFO
	WARNING
	ERROR
	CRITICAL
)

// A Logger with a null impl pointer is effectively a root logger.
type Logger struct {
	impl *module
}

type module struct {
	name   string
	level  Level
	parent *module
}

// Initially the modules map only contains the root module.
var (
	root    = &module{"", WARNING, nil}
	modules = map[string]*module{
		"": root,
	}
)

func (level Level) String() string {
	switch level {
	case NOT_SPECIFIED:
		return "NOT_SPECIFIED"
	case TRACE:
		return "TRACE"
	case DEBUG:
		return "DEBUG"
	case INFO:
		return "INFO"
	case WARNING:
		return "WARNING"
	case ERROR:
		return "ERROR"
	case CRITICAL:
		return "CRITICAL"
	}
	return "<unknown>"
}

func GetLogger(name string) Logger {
	// Lowercase the module name, and look for it in the modules map.
	name = strings.ToLower(name)
	impl, found := modules[name]
	if found {
		return Logger{impl}
	}

	parent := root
	// Really wanting rsplit.
	parts := strings.Split(name, ".")
	if len(parts) > 1 {
		parent = GetLogger(strings.Join(parts[:len(parts)-1], ".")).impl
	}
	impl = &module{name, NOT_SPECIFIED, parent}
	modules[name] = impl
	return Logger{impl}
}

func DumpLoggingLevels() string {
	output := []string{}
	// output in alphabetical order.
	keys := []string{}
	for key := range modules {
		keys = append(keys, key)
	}
	sort.Strings(keys)
	for _, name := range keys {
		mod := modules[name]
		severity := mod.level
		if severity == NOT_SPECIFIED {
			continue
		}
		module_name := mod.name
		if module_name == "" {
			module_name = "<root>"
		}
		output = append(output, fmt.Sprintf("%s %s", module_name, severity))
	}
	return strings.Join(output, "\n")
}

func ConfigureLogging(configString string) {
	if configString == "" {
		return
	}

	values := strings.FieldsFunc(configString, func(r rune) bool { return r == ';' || r == ':' })
	for _, value := range values {
		s := strings.SplitN(value, "=", 2)
		if len(s) < 2 {
			continue
		}
		name := s[0]
		level := s[1]
		if name == "<root>" {
			name = ""
		}
		logger := GetLogger(name)
		logger.SetLogLevel(GetLoggingLevel(level))
	}
}

// GetLoggingLevel converts a string representation of a logging level to the
// enumerated value.
func GetLoggingLevel(level string) Level {
	level = strings.ToUpper(level)
	switch level {
	case "TRACE":
		return TRACE
	case "DEBUG":
		return DEBUG
	case "INFO":
		return INFO
	case "WARN", "WARNING":
		return WARNING
	case "ERROR":
		return ERROR
	}
	return WARNING
}

func (logger *Logger) getModule() *module {
	if logger == nil || logger.impl == nil {
		return root
	}
	return logger.impl
}

func (logger *Logger) Module() string {
	return logger.getModule().name
}

func (logger *Logger) GetLogLevel() Level {
	return logger.getModule().level
}

func (module *module) getEffectiveLogLevel() Level {
	if module.level == NOT_SPECIFIED && module.parent != nil {
		return module.parent.getEffectiveLogLevel()
	}
	return module.level
}

func (logger *Logger) GetEffectiveLogLevel() Level {
	return logger.getModule().getEffectiveLogLevel()
}

func (logger *Logger) SetLogLevel(level Level) {
	module := logger.getModule()
	// The root module can't be unspecified.
	if module.name == "" && level == NOT_SPECIFIED {
		level = WARNING
	}
	module.level = level
}

func (logger *Logger) log(level Level, message string, args ...interface{}) {
	// Gather time, and filename, line number.
	now := time.Now() // get this early.
	var file string
	var line int
	// Param to Caller is the call depth.  Since this method is called from
	// the Logger methods, we want the place that those were called from.
	_, file, line, ok := runtime.Caller(2)
	if !ok {
		file = "???"
		line = 0
	}

	formattedMessage := fmt.Sprintf(message, args...)
	writeToWriters(level, logger.impl.name, file, line, now, formattedMessage)
}

func (logger *Logger) Critical(message string, args ...interface{}) {
	logger.log(CRITICAL, message, args...)
}

func (logger *Logger) Error(message string, args ...interface{}) {
	if logger.IsErrorEnabled() {
		logger.log(ERROR, message, args...)
	}
}

func (logger *Logger) Warning(message string, args ...interface{}) {
	if logger.IsWarningEnabled() {
		logger.log(WARNING, message, args...)
	}
}

func (logger *Logger) Info(message string, args ...interface{}) {
	if logger.IsInfoEnabled() {
		logger.log(INFO, message, args...)
	}
}

func (logger *Logger) Debug(message string, args ...interface{}) {
	if logger.IsDebugEnabled() {
		logger.log(DEBUG, message, args...)
	}
}

func (logger *Logger) Trace(message string, args ...interface{}) {
	if logger.IsTraceEnabled() {
		logger.log(TRACE, message, args...)
	}
}

func (logger *Logger) IsErrorEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= ERROR
}

func (logger *Logger) IsWarningEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= WARNING
}

func (logger *Logger) IsInfoEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= INFO
}

func (logger *Logger) IsDebugEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= DEBUG
}

func (logger *Logger) IsTraceEnabled() bool {
	return logger.getModule().getEffectiveLogLevel() <= TRACE
}
